"""logging

logging module add mainly utils for logThread and @log decorator to log function calles"""

from jpe_types.logging import logUtils, stepLogging
from jpe_types.logging.threadLogging import log, _setupLoggers
from jpe_types.logging.stepLogging import log as stepLog
from os import system
from os.path import dirname, isdir
from shutil import rmtree, copyfile

__all__ = ["setup", "openConfig", "log", "logUtils", "clearLog", "stepLog"]

def setup(runName=None):
    """setup the thread loggers
    
    @param runName: overide to std run name if None default us std values set in loggingConfig.json
    @type runName: str or None"""
    assert isinstance(runName, str) or runName is None, f"runName must be str or None not {type(runName)}"
    setattr(logUtils, "runFile", logUtils.create_newRun(runName))

    _setupLoggers()

def clearLog(file: str="./frameLog")->None:
    """remove log dir aka the same dir as the one passed to log

        @param file: the name of the logg tree
        @type file: str
        """
    if not type(file) is str: raise TypeError(f"param file must be str not {type(file)}")
    if isdir(file):
        rmtree(file)

def openConfig():
    """opens the config.json file
    """
    system(f"{dirname(__file__)}/loggingConfig.json")
    val = "n"
    while val!="y":
        val = input("are you done changing confing ?\n(y/n)")

    copyfile(f"{dirname(__file__)}/loggingConfig.json", 
             f"{dirname(__file__)}/loggingConfig.archive.json")
    


from jpe_types.utils import copy


def combineDicts(master: dict, lower: dict):
    """combine 2 dicts

    dot this later
    """
    res = copy(master)
    for key, val in zip(lower, lower.values()):
        if isinstance(val, dict) and key in res:
            res[key] = combineDicts(res[key], lower[key])
            continue
        res[key] = copy(val)

    
    
    return res


# distutils: language = c++

from libcpp.string cimport string as cstr
from libcpp.list cimport list as clist


cdef extern from "jpe_typeUtilsExtentions/FileScanner.h":
    clist[int] findRefrences(cstr, cstr)

cpdef list[int] CppfindFalsInFile(bytes fileName_py, bytes val_py):

    cdef cstr fileName = fileName_py
    cdef cstr val = val_py

    cdef clist[int] res = findRefrences(fileName, val)
    return res




"patch fore coping problems"

import copy as copyLib
    
__all__ = ["copy"]

def copy(val):
    """crates a copy of a list or dict

    this is a fix for the problem that when coping a list it of a list it dosnt copy second dim list

        >>> a = [[1,2],[3,4]]
        >>> b = copy.copy(a)
        # std copy lib copy 

        >>> a[0] = 10
        # in this case b is [[1,2],[3,4]]
        # so it works 

    however

        >>> a[1][0] = 10
        # in this case b is [[1,2],[10,4]] but in a true copy it schould be [[1,2], [3,4]]
        # if we use this patch
        >>> a = [[1,2],[3,4]]
        >>> c = jpe.utils.copy.copy(a)
        >>> a[0][0] = 10
        >>> b
        [[1,2],[3,4]]

    @param val: object what we want to copy
    @type val: list, dict, anything else will be done via copy.copy

    @return: pyobject
            a true copy of val

    """
    if isinstance(val, list):
        output = []
        for element in val:
            output.append(copy(element))
        return output

    elif isinstance(val, dict):
        output ={}
        for element in val:
            output[copy(element)] = copy(val[element])
        return output

    return copyLib.copy(val)
    



"""utils

logg utils function akka log builders and config handeling kida
"""

import logging, json, os, typing
from threading import current_thread
from datetime import datetime
#from jpe_types.paralel.thread import threadInharitanceFilter

logging_data = json.load(open(f"{os.path.dirname(os.path.realpath(__file__))}/loggingConfig.json","r"))
"config dict"

handlerDict: typing.Dict = {
    #"BaseRotatingHandler": logging.handlers.BaseRotatingHandler,
    #"RotatingFileHandler": logging.handlers.RotatingFileHandler,
    #"TimedRotatingFileHandler": logging.handlers.TimedRotatingFileHandler,
    #"WatchedFileHandler": logging.handlers.WatchedFileHandler,
    #"SocketHandler": logging.handlers.SocketHandler,
    #"DatagramHandler": logging.handlers.DatagramHandler,
    #"SysLogHandler": logging.handlers.SysLogHandler,
    #"SMTPHandler": logging.handlers.SMTPHandler,
    #"NTEventLogHandler": logging.handlers.NTEventLogHandler,
    #"HTTPHandler": logging.handlers.HTTPHandler,
    #"BufferingHandler": logging.handlers.BufferingHandler,
    #"MemoryHandler": logging.handlers.MemoryHandler,
    #"QueueHandler": logging.handlers.QueueHandler,
    
    "fileHandler": logging.FileHandler,
    "StreamHandler": logging.StreamHandler}
"handlr str to class dict"

def create_newRun(name=None):
    """create new run

        don't use this function unless you want to mess stuff up
        
        name generation in loggingConfig.json

        @param id: the number of the run

        %a Sun, Mon, ...
        %A Full weekday name.
        %w Weekday as a decimal number.
        %d Day of the month as a zero-padded decimal.
        %-d Day of the month as a decimal number.
        %b Abbreviated month name.
        %B Full month name.
        %m Month as a zero-padded decimal number.
        %-m Month as a decimal number.
        %y Year without century as a zero-padded decimal number.
        %-y Year without century as a decimal number.
        %Y Year with century as a decimal number.
        %H Hour (24-hour clock) as a zero-padded decimal number.
        %-H Hour (24-hour clock) as a decimal number.
        %I Hour (12-hour clock) as a zero-padded decimal number.
        %-I	Hour (12-hour clock) as a decimal number.
        %p Locale’s AM or PM.
        %M Minute as a zero-padded decimal number.
        %-M Minute as a decimal number.
        %S Second as a zero-padded decimal number.
        %-S Second as a decimal number.
        %f Microsecond as a decimal number, zero-padded on the left.
        %z UTC offset in the form +HHMM or -HHMM.
        %Z Time zone name.
        %j Day of the year as a zero-padded decimal number.
        %-j Day of the year as a decimal number.
        %U Week number of the year (Sunday as the first day of the week). All days in a new year preceding the first Sunday are considered to be in week 0.
        %W Week number of the year (Monday as the first day of the week). All days in a new year preceding the first Monday are considered to be in week 0.
        %c Locale’s appropriate date and time representation.
        %x Locale’s appropriate date representation.
        %X Locale’s appropriate time representation.
        %% A literal '%' character.
    
        see https://www.programiz.com/python-programming/datetime/strftime for extra detailt ofn datetimeing"""
    if name is None:
        name = str(logging_data["logDirName"])

    zeros = logging_data["logZeroFill"]

    def getCustomName(name):
        timeStamp = datetime.now()
        name = timeStamp.strftime(name)
        nname = name.format(id=run_number)
        return nname

    getPath = lambda run_number: f"./log/{getCustomName(name)}" if run_number==0 else f"./log/{getCustomName(name) + str(run_number)}"
    
    if not os.path.isdir("./log"):
        os.makedirs("./log")

    run_number = 0
    path = getPath(run_number)
    while os.path.isdir(path):
        run_number += 1
        path = getPath(run_number)
    
    os.makedirs(path)

    return path +"/"

runFile = None
"the file to runn to"


def Create_defaultLogger(name:str = current_thread().name, logType:str="stdLogger")  ->  typing.NewType("logger", logging.Logger):
    """create std logger
    
    create std logger as specified by loggingConfig.json so edit that to change it

    @param name: the name of the logger
    @type name: str

    @param logType: the type of a logger as set by loggingConfig
    @type logType: str
    """
    if runFile is None:
        raise LogSetupError()

    FORMAT = logging_data[logType]["Format"]

    log = logging.getLogger(name)

    handler = handlerDict[logging_data[logType]["handler"]]
    syslog = handler(f"{runFile}{name}.log")

    #log.addFilter(threadInharitanceFilter())

    formatter = logging.Formatter(FORMAT)
    syslog.setFormatter(formatter)
    
    log.setLevel(logging_data[logType]["logLevel"])
    log.addHandler(syslog)

    log.setLevel(logging_data[logType]["logLevel"])
    for msg in logging_data[logType]["startupConfig"]:
        log.info(msg)

    return log


class LogSetupError(Exception):
    "Exception raised when logging wasnt setup befor calls"
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, f"logging must be initiated but want run jpe_types.logging.setup() before using loggers", **kwargs)



"setup script"

import json
from os.path import dirname, join
from os import system
from sys import executable

def run():
    #update logg utils json file

    path = dirname(__file__)
    #print( f"{executable} {join(path, u"utils/Optimizations/build.py")}")
    buildScript_path = join(path, "utils", "Optimizations", "build.py")
    outputScriptPath = join(path, "utils", "Optimizations")
    system(f"{executable} {buildScript_path} build_ext --build-lib {outputScriptPath}")

if __name__ == "__main__":
    run()


"""compleatly logg function execution

    log function execution line by line
    its threadsafe btw
    """

import sys, time
from threading import current_thread, get_ident, Thread
import traceback, typing
from os.path import samefile, isdir, join
from os import mkdir, listdir
from pathlib import Path
from jpe_types.conversions.intager import baseN
from pickle import dump, load
from tkinter import Tk, Button, Text, END, Label, Scale, HORIZONTAL
from typing import NewType
from inspect import getsource, getsourcelines, getsourcefile
from shutil import rmtree


__all__ = ["SetTrace", "FuncFrames", "loggers", "log", "replay"]
"emmm"

def encriptName(value: int)->str:
    """encript frame id to save 

        it converts a value int to base 36 
        """
    #return baseN(value, 97, numerals="1234567890'^qwertzuiopasdfghjkl$yxcvbnm,-+ç%&/()=`QWERTZUIOP!ASDFGHJKLYXCVBNM;:_<>¦@#°§¬|¢}]~[{£¨")
    return baseN(value, 36, numerals="1234567890qwertzuiopasdfghjklyxcvbnm")
    #return hex(value)

class SetTrace(list):
    """internal
    with SetTrace(monitor):

    the Tracer used to combine everything hapeening in a certain Thread.
    """
    funcCode=""
    lastFrame = None
    startFile = None
    def __init__(self, func, name, tb=[],  directory ="./frameLog"):
        """constructor

            create the listener witch will call func at every log call made by sys
            the name is equivelent to the directory name to log to

            @param func: the logging function
            @type func: function

            @param name: the name of the Thread to that is being loged or more accuratly the name of the subdir in dir to log to
            @type name: str

            @param tb: the starting tb of the log
            @type tb: list<str>

            @param directory: the superdir to log to defualts to ./frameLog
            @type direcroy: os pathname
            """

        assert hasattr(func, "__call__"), f"param func must be callable"
        self.func = func(self)
        "the tracing function"
        assert hasattr(name, "__str__"), f"param name must be str not {type(name)}"
        self.name = str(name)
        "the name of the thread it runs on"
        self._add_active = False
        "weather or not a new function trace is added"
        self._active = False
        "weather its acrive"
        self._name = current_thread().name
        "name of the thread that started it [depricated]"
        self.key = 0
        "the howmaniath log we loged cant be larger than 36**31"
        if not hasattr(tb, "__iter__"): raise TypeError(f"param tb must be ")
        list.__init__(self, tb)
        #self.mainStack = mainStack

        if not hasattr(directory, "__str__"): raise TypeError(f"param directory must be str not {type(directory)}")
        self._dir = directory
        """the directory to log to
        may not be changed"""

        if not isdir(self._dir):
            mkdir(self._dir)
        
        if not isdir(f"{self._dir}/{self._name}"):
            mkdir(f"{self._dir}/{self._name}")
    
    def __str__(self):
        return f"debug Tracer with tb {list.__str__(self)}"
    
    def append(self, val)->None: 
        """append value

        append the value to the tracers tb list if its active"""
        if not samefile(str(val[-1])[19:str(val[-1]).index(",")], Path(__file__)):
            if self._add_active and not val in self:
                list.append(self, self._formatTb(val))
            self._add_active = False
    
    def __contains__(self, other):
        """weather a tb schould be loged
        """
        return list.__contains__(self, self._formatTb(other))
    
    @staticmethod
    def _formatTb(val):
        "get file location"
        tb = []
        
        for x in val:
            x = str(x)
            kommaPos = x.index(",")
            y = x[kommaPos+5:]
            inPos = y.index("in")
            tb.append(x[19:kommaPos] + y[inPos+2:])
        return tb

    def start(self):
        "start the log"
        if not self._active:
            self._active = True
            sys.settrace(self.func)
    def __enter__(self):
        return self
    def __exit__(self, ext_type, exc_value, traceback):
        self._active = False
        sys.settrace(None)
        # http://effbot.org/zone/python-with-statement.htm
        # When __exit__ returns True, the exception is swallowed.
        # When __exit__ returns False, the exception is reraised.
        # This catches Sentinel, and lets other errors through
        # return isinstance(exc_value, Exception)

class FuncFrames(list):
    """the current Frame
    
        a class used to extract and save data from the sys frame """
    def __init__(self, frame, event, arg, Tracer):
        """constructor

            creatae the frame and extract data

            @param frame: the sys frame to extract from
                            !no assersions
            @type frame: sts.Frame

            @param event: what type of event hapend
            @type event: str

            @param args: the arg param passed by sys
            @type arg: ???

            @param Trace: the trace that this frame belogs to 
            @type Trace: SetTrace class defined here
            """
        assert type(event) is str, f"param event must be str"

        def getLocals():
            res = {}
            for key, val in frame.f_locals.items():
                if hasattr(val, "__str__"):
                    res[key] = str(val)
            return res


        self.local: dict = getLocals()
        self.event = event
        self.arg   = arg
        self.func = frame.f_code.co_name
        self.funcCode = getsource(frame.f_code)
        self.lineNumber = frame.f_lineno - getsourcelines(frame.f_code)[1]+1
        self.startedThreads = []
        self.endedThreads = []

        self.key   = Tracer.key
        self.trace = Tracer
        Tracer.key += 1
        Tracer.lastFrame = self
    
    def __str__(self):return str(self.startedThreads)
    
    def save(self):
        "save the state"
        dump(self, open(f"{self.trace._dir}/{self.trace._name}/{encriptName(self.key)}.plog", "wb"))
    
    def __getstate__(self):
        value={"local": self.local,
                "event": self.event,
                "arg": str(self.arg) if hasattr(self.arg, "__str__") else "sry could not str convert",
                "func": self.func,
                "funcCode": self.funcCode,
                "lineNumber": self.lineNumber,
                "startedThreads": self.startedThreads,
                "endedThreads": self.endedThreads}
        
        return value
            
    def __setstate__(self, state):
        def load(local, event, arg, func, funcCode, lineNumber, startedThreads, endedThreads):
            self.local = local
            self.event = event
            self.arg = arg
            self.func = func
            self.funcCode = funcCode
            self.lineNumber = lineNumber
            self.startedThreads = startedThreads
            self.endedThreads = endedThreads
        load(**state)
funcFramesType = NewType("FuncFrames", FuncFrames)
"typing type of FuncFrames"

def execLog(Trace: typing.NewType("Trace", SetTrace)):
    """logging function used ty log decorator
    
        purly internal"""
    def monitor(frame, event, arg):
        if Trace._active:
            tb = traceback.extract_stack(f=frame)
            if len(tb) == 0:
                return
            if samefile(str(tb[-1])[19:str(tb[-1]).index(",")], Path(__file__)):
                return
            Trace.append(tb)
            if tb in Trace:
                if event == "line" or event == "return":
                    saveFrame = FuncFrames(frame, event, arg, Trace)
                    saveFrame.save()

                    
        return monitor
    return monitor

loggers: typing.Dict[str, typing.Dict[Thread, SetTrace]] = {}
"""dict of loggers

    a dict conaining loggers by save file and thread it contains SetTrace's

    @type: dict<str, dict<Thread, SetTrace>>
    """
logFile: typing.Dict[Thread, str] = {}
"""dict of files by THread

    the files in witch the log is saved by Thread

    @type: dict<Thread, str>
    """

def log(directory: str="./frameLog"):
    """logging decorator

        log the functions this relates to

        @param directory: the directory to save to
        @type directory: str
        """
    def sublog(function):
        def newLogger(args, kwargs):
            "if logger dosent exist create a new one"
            with SetTrace(execLog, current_thread().name, directory=directory) as Trace:
                #add logger dir
                if not directory in loggers:
                    loggers[directory] = {}

                loggers[directory][current_thread()] = Trace
                Trace.startFile = getsourcefile(function)
                return mainLog(Trace, args, kwargs, function) 
     
        def reuseLogger(args, kwargs):
            "if logger existr reuse it"
            Trace = loggers[directory][current_thread()]
            return mainLog(Trace, args, kwargs, function)       
        def mainLog(Trace, args, kwargs, func):
            "run function"
            logFile[current_thread()] = directory
            Trace.start()
            Trace._add_active = True
            ret = func(*args, **kwargs)
            return ret
            
        def main(*args, **kwargs):
            "fingure out witch function schuld be used"
            if directory in loggers:
                if current_thread() in loggers[directory]: 
                    return reuseLogger(args, kwargs)
            return newLogger(args, kwargs)
            

        return main
    return sublog






class replay(Tk):
    """replay a log file

        replay logg file
        """
    def __init__(self, file: str="./frameLog", name: str="MainThread", baseTime: float=10, master=None):
        """constructor

            @param file: the dir in witch stuff got loged
            @type file: str
            @param name: the thread name to be used
            @type name: str
            @param baseTime: the max time to wait beween auto steps
            @type baseTime: float
            @param master: the log of the thread that initiated it
            @type master: replay
            """
        assert isinstance(file, str), f"param file must be str not {type(file)}"
        self._dir:str = file
        "the dir of the log"
        assert isinstance(name, str), f"param name must be str not {type(name)}"
        self.name:str = name
        "the threads name"
        Tk.__init__(self)
        self.baseTime: float = baseTime
        "max time betwean steps"
        self.master = self if master is None else master
        self.kids: list = [self]
        "every instance of reply with this as tis master"
        self.kidThreads = {}
        "the threads of all threads used by master"

        self._autoStep=False

        self.key: int = 0

        self._createTk()
    
    def autoStep(self):
        while self._autoStep:
            if self.stepThrew.get()==0:
                continue

            time.sleep(abs(self.baseTime/self.stepThrew.get()))
            for element in self.kids:
                if self.stepThrew.get()>0:
                    element.next()
                else:
                    element.prev()

    def toggleAutoSep(self):
        params = {True: "stop autostep",
                  False:"start autostep"}
        self.master._autoStep = not self.master._autoStep
        for element in self.master.kids:
            element.autoStepThrewCode.config(text=params[self.master._autoStep])
        if self.master._autoStep:
            Thread(target=self.master.autoStep,
                   daemon=True,
                   name="stepThrewCode").start()

    def _createTk(self):
        self.codeText = Text(self) 
        self.codeText.pack()

        self.var_label = Text(self)
        self.var_label.pack()
        self.title(f"Thread {self.name}")

        self.nextButton = Button(self,
                                 command=self.next,
                                 text="next")
        self.nextButton.pack(side="bottom")

        self.prevButton = Button(self,
                                 command=self.prev,
                                 text="last")
        self.prevButton.pack(side="bottom")

        if self.master is self:
            self.stepThrew = Scale(self, from_=-100, to=100, orient=HORIZONTAL)
            self.stepThrew.pack()

        self.autoStepThrewCode = Button(self, command=self.master.toggleAutoSep, text="start autostep")
        self.autoStepThrewCode.pack()

        self.show()
    
    def show(self) -> None:
        "show the log state"
        data: funcFramesType = load(open(f"{self._dir}/{self.name}/{encriptName(self.key)}.plog", "rb"))

        def create_Thread(threadName):
            newThreadLog = replay(self._dir, threadName, master=self.master)
            self.master.kids.append(newThreadLog)
            self.master.kidThreads[threadName] = newThreadLog
            newThreadLog.mainloop()

        for threadName in data.startedThreads:
            if isdir(join(self._dir, threadName)):
                tlogExe = Thread(target=create_Thread,
                                args=(threadName,),
                                daemon=True,
                                name=f"replay {threadName}")
                tlogExe.start()
        
        for removedThreads in data.endedThreads:
            if removedThreads in self.master.kidThreads:
                self.master.kidThreads[removedThreads].destroy()
            

        self.var_label.delete("1.0", END)
        self.codeText. delete("1.0", END)
        for key, val in data.local.items():
            self.var_label.insert(END, f"{key} is {val}\n")
        
        if data.event == "return":
            self.var_label.insert(END, f"return Value is {data.arg}")


        self.codeText.insert(END, data.funcCode)
        self.codeText.tag_add("currentLine", f"{data.lineNumber}.0", f"{data.lineNumber}.end")
        self.codeText.tag_config("currentLine", background="yellow", foreground="black")


    def __iadd__(self, val):
        assert isinstance(val, int), f"cant use non int stepsize"
        if self.key+val < 0: 
            self.key=0
        elif f"{encriptName(self.key+val)}.plog" in listdir(f"{self._dir}/{self.name}"):
            self.key += val
        self.show()
    def __isub__(self, val):
        self.__iadd__(-val)
    def next(self): self.__iadd__(1)
    def prev(self): self.__isub__(1)




"better types for python"
from jpe_types import paralel
from jpe_types import conversions
from jpe_types import logging
from jpe_types import utils
from jpe_types import caching

__all__=["paralel", "conversions", "logging", "utils", "caching"]


"""cach dir 

see chach dir"""
from jpe_types.caching.cachDirPy import strCachDir
from shutil import rmtree
from os.path import isdir

__all__ = ["strCachDir", "clearCach"]

def clearCach():
    "cleat data cach"
    if isdir("./cach"):
        rmtree("./cach")


import os, typing
from warnings import warn
try: 
    from jpe_types.utils.Optimizations.subScripts import CppfindFalsInFile
except ImportError:
    warn("jpe_utils.utils.Scanning Cython lod Failure")
    import jpe_types.setup
    jpe_types.setup.run()
    CppfindFalsInFile = None
        
class filePosition:
    """the location of a val in files

        the position in the a file set as file with line number created by search algorythem
    """
    def __init__(self, file: str, line: int, doAssersions: bool=True):
        """constructor

            create a new instance of the filePosition containing file and line

            @param file: the file in witch the val was found
            @type file: os path string of a file

            @param line: the line in witch the value was found
            @type line: int

            @param doAssesions: weather or not to assert values
            """

        if doAssersions:
            assert isinstance(file, str), f"file must be str not{type(file)}"
            assert os.path.isfile(file), f"file must be a valid file, {file} is not a valid path"
            assert isinstance(line, int), f"line must be int not {type(line)}"

        self.file = file
        "the fill in witch a solution was found"

        self.line = line   
        "the line number the solution was found in"

    def __str__(self) -> str:
        return f"in file {self.file} at line {self.line}"

    __repr__ = __str__
filePositionType = typing.NewType("filePosition", filePosition)

class GenralFilesPostions(list):
    """the sumary of positions of filePositions

        a subclass of list with all its functionality howerver it adds the posibility to 
            - get the files in witch calles accured as set
    """
    def getFiles(self) -> set:
        """get all the files a result was found in

            @return: set of the paths of the files
        """
        res = set()
        for pos in self:
            res.add(pos.file)
        return res
    
    def __str__(self):
        res = "found solutions here:"
        for value in self:
            res +=  "\n" + value.__str__()
        return res
    def getFilePoses(self):
        """get all the files with line number a result was found in

            @return: set of tuples(file path, line number)
        """
        res = set()
        for pos in self:
            res.add((pos.file, pos.line))
        return res
GenralFilesPostionsType = typing.NewType("GenralFilesPostions", GenralFilesPostions)

def scanLog(val:str ,root: str="./log") -> GenralFilesPostionsType:
    """ask theo
    """
    positions = GenralFilesPostions()
    for path in os.listdir(path=root):
        
        new_path = os.path.join(root, path)
        if os.path.isdir(new_path):
            positions += scanLog(val, new_path)
        else:
            if not CppfindFalsInFile is None:
                positions += cyScanLogFile(val, new_path)
            else:
                positions += pyScanLogFile(val, new_path)
    
    return positions

def pyScanLogFile(val: str, path: str) -> GenralFilesPostionsType:
    "subrutine dont use"
    
    positions = GenralFilesPostions()
    with open(path) as f:
        logFile = f.readlines()
        for lineNumber, line in enumerate(logFile):
            if val in line:
                positions.append(filePosition(path, lineNumber))
    
    return positions
def cyScanLogFile(val: str, path: str) -> GenralFilesPostionsType:
    "subrutine dont use"
    positions = GenralFilesPostions()

    posses = CppfindFalsInFile(path.encode(), val.encode())
    for val in posses:
        positions.append(filePosition(path, val))
    
    return positions



"better Threads and Thread logging"
import threading
import logging
import typing
from traceback import format_exc
from jpe_types.logging.logUtils import Create_defaultLogger, logging_data, LogSetupError
from jpe_types.logging.threadLogging import main_Logger, error_Logger, _setupLoggers
from jpe_types.logging.stepLogging import loggers as stepLoggers, logFile as stepLogFiles 


class Thread(threading.Thread):
    """an extended version of the std python thread

        see https://docs.python.org/3/library/threading.html for more information on threads

        but dont sublass the run method or the return data will be lost, instead  replace run with execute

        this threads adds
            - thread inheritance so we can track how the thread was creted mostly for debugging
            - thread returns the ability to return a tupple form a thread
        """
    def __init__(self, group=None, target=None, name=None, args=(), kwargs=None, *, daemon=None):
        """This constructor should always be called with keyword arguments. Arguments are:

        *group* should be None; reserved for future extension when a ThreadGroup
        class is implemented.

        *target* is the callable object to be invoked by the run()
        method. Defaults to None, meaning nothing is called.

        *name* is the thread name. By default, a unique name is constructed of
        the form "Thread-N" where N is a small decimal number.

        *args* is the argument tuple for the target invocation. Defaults to ().

        *kwargs* is a dictionary of keyword arguments for the target
        invocation. Defaults to {}.

        If a subclass overrides the constructor, it must make sure to invoke
        the base class constructor (Thread.__init__()) before doing anything
        else to the thread.

        it alsow alowes the step logging setup to log open the threads log file if it exists

        """
        threading.Thread.__init__(self, group=group, target=target, name=name, args=args, kwargs=kwargs, daemon=daemon)

        self.parent = threading.current_thread()
        """witch tread initiated this thread

        the thread, that initiated this one, for inharitance.
        currently mostly for logging
        """

        self._returnVal = None
        """the result of the treads Function

        used to store the return value of the function the thread ran
        unill join is called at witch point return val is returned
        """

        self._traceDir = None
        """direcory of the log

        the directory of the log file"""

    def start(self):
        """Start the thread’s activity.

            It must be called at most once per thread object. It arranges for the object’s run() method to be invoked in a separate thread of control.
            This method will raise a RuntimeError if called more than once on the same thread object."""
        threading.Thread.start(self)

    def run(self):
        """runs the thread and saves the result"""
        self._returnVal = self.execute()    
    
    def execute(self):
        """Method representing the thread's activity.

            You may override this method in a subclass. The standard run() method
            invokes the callable object passed to the object's constructor as the
            target argument, if any, with sequential and keyword arguments taken
            from the args and kwargs arguments, respectively."""
        try:
            if self._target:
                val = self._target(*self._args, **self._kwargs)
        finally:
            # Avoid a refcycle if the thread is running a function with
            # an argument that has a member that points to the thread.
            del self._target, self._args, self._kwargs
        return val
    
    def join(self):
        """Wait until the thread terminates.

            This blocks the calling thread until the thread whose join() method is
            called terminates -- either normally or through an unhandled exception
            or until the optional timeout occurs.

            When the timeout argument is present and not None, it should be a
            floating point number specifying a timeout for the operation in seconds
            (or fractions thereof). As join() always returns None, you must call
            isAlive() after join() to decide whether a timeout happened -- if the
            thread is still alive, the join() call timed out.

            When the timeout argument is not present or None, the operation will
            block until the thread terminates.

            A thread can be join()ed many times.

            join() raises a RuntimeError if an attempt is made to join the current
            thread as that would cause a deadlock. It is also an error to join() a
            thread before it has been started and attempts to do so raises the same
            exception."""

        threading.Thread.join(self)
        if self._traceDir in stepLoggers:
            if threading.current_thread() in stepLoggers[self._traceDir]:
                frame = stepLoggers[self._traceDir][threading.current_thread()].lastFrame
                if not frame is None:
                    frame.endedThreads.append(self.name)
                    frame.save()

        return self._returnVal

class LockableThread(Thread):
    """a combination of a lock and a thread

        makes it easyer to keep track of locks and threads and unlock, lock them
        """
    def __init__(self, group=None, target=None, name=None, args=(), kwargs=None, lock=threading.Lock, *, daemon=None):
        """This constructor should always be called with keyword arguments. Arguments are:

            @param group: should be None; reserved for future extension when a ThreadGroup
            class is implemented.

            @param target: is the callable object to be invoked by the run()
            method. Defaults to None, meaning nothing is called.

            @param name: is the thread name. By default, a unique name is constructed of
            the form "Thread-N" where N is a small decimal number.

            @param args: is the argument tuple for the target invocation. Defaults to ().

            @param kwargs: is a dictionary of keyword arguments for the target
            invocation. Defaults to {}.

            If a subclass overrides the constructor, it must make sure to invoke
            the base class constructor (Thread.__init__()) before doing anything
            else to the thread.

            """
        Thread.__init__(self, group=group, target=target, name=name, args=args, kwargs=kwargs, daemon=daemon)

        self.lock = lock()
        "the lock used to lock the thread"
    
    def aquire(self, blocking=True, timeout=-1):
        """Acquire a lock, blocking or non-blocking.

        When invoked without arguments: if this thread already owns the lock,
        increment the recursion level by one, and return immediately. Otherwise,
        if another thread owns the lock, block until the lock is unlocked. Once
        the lock is unlocked (not owned by any thread), then grab ownership, set
        the recursion level to one, and return. If more than one thread is
        blocked waiting until the lock is unlocked, only one at a time will be
        able to grab ownership of the lock. There is no return value in this
        case.

        When invoked with the blocking argument set to true, do the same thing
        as when called without arguments, and return true.

        When invoked with the blocking argument set to false, do not block. If a
        call without an argument would block, return false immediately;
        otherwise, do the same thing as when called without arguments, and
        return true.

        When invoked with the floating-point timeout argument set to a positive
        value, block for at most the number of seconds specified by timeout
        and as long as the lock cannot be acquired.  Return true if the lock has
        been acquired, false if the timeout has elapsed.

        """
        self.lock.acquire(blocking=blocking, timeout=timeout)
    
    def releace(self):
        """Release a lock, decrementing the recursion level.

            If after the decrement it is zero, reset the lock to unlocked (not owned
            by any thread), and if any other threads are blocked waiting for the
            lock to become unlocked, allow exactly one of them to proceed. If after
            the decrement the recursion level is still nonzero, the lock remains
            locked and owned by the calling thread.

            Only call this method when the calling thread owns the lock. A
            RuntimeError is raised if this method is called when the lock is
            unlocked.

            There is no return value.

            """
        self.lock.release()

class LogThread(LockableThread):
    """a thread wit its owne logger

        a thread with a logger to log all acivity of the logger using decorators
        """    
    threadLogger = None
    def __init__(self, *args, **kwargs):
        """constructor
        
            see LockableThread.__init__"""
        LockableThread.__init__(self, *args, **kwargs)
        global main_Logger, error_Logger
        # if error_Logger is None: _setupLoggers()
        if main_Logger is None:
            from jpe_types.logging.threadLogging import main_Logger, error_Logger

        self.threadLogger = Create_defaultLogger(name=self.name, logType="threadLogger")

    def start(self):
        """Start the thread’s activity.

            It must be called at most once per thread object. It arranges for the object’s run() method to be invoked in a separate thread of control.
            This method will raise a RuntimeError if called more than once on the same thread object.
            and make log call"""
        
        self.threadLogger.info(f"created thread {self.name}")
        LockableThread.start(self)


    def set_logger(self, log):
        """sets the logger

            set the logger of the thead used as an overide to std log used by the logging setup
            """
        self.threadLogger = log
    
    def run(self):
        """runs the thread and saves the result

            run the exec and logg the resut if error
            """
        try:
            LockableThread.run(self)

        except Exception as e:
            # logg on additional loggers
            for subLoggerStr in logging_data["threadLogger"]["additionalLoggers"]:
                #log with and without traceback
                if logging_data[logging_data["logTranslator"][subLoggerStr]]["trackBack"]:
                    eval(subLoggerStr).error(format_exc())
                else:
                    eval(subLoggerStr).error(e)
                #logg end msg
                for msg in logging_data["threadLogger"]["errorLogMessage"]:
                    eval(subLoggerStr).error(msg)
            
            if logging_data["threadLogger"]["logErrors"]:
                if logging_data["threadLogger"]["trackBack"]:
                    self.threadLogger.error(format_exc())
                else:
                    self.threadLogger.error(e)

                for msg in logging_data["threadLogger"]["errorLogMessage"]:
                    self.threadLogger.error(msg)
            
            raise e
        
        return self._returnVal

class threadInharitanceFilter(logging.Filter):
    """logging filter for thread inheritance
    
        thread inheritance filter

        adds the ability to incorporate the inheritance of the thread from witch the log call originated
        like this
        
        >>> log = logging.getLogger(__name__)

        >>> syslog = logging.FileHandler("app.log")
        >>> log.addFilter(jpe_types.paralel.threadInharitanceFilter())

        >>> formatter = logging.Formatter('%(asctime)s in thread %(threadInharitace)s : %(message)s')
        >>> syslog.setFormatter(formatter)
        >>> log.setLevel(logging.DEBUG)
        >>> log.addHandler(syslog)
        """
    def filter(self, record):
        """generate str or thred inheritance
        
            crates a str containing the inheritance of the thread that called this function
            
            the str will be of type
            parentName.thisThreadName 
            
            add to the same for the parent etc"""
        record.threadInharitace = threadInharitanceFilter.getThreadName(threading.current_thread())
        return True
    
    @staticmethod
    def getThreadName(thread: threading.Thread):
        "recursivly compute the thread inheritance string"
        if isinstance(thread, Thread):
            return threadInharitanceFilter.getThreadName(thread.parent) +"."+thread.name
        return thread.name



from setuptools import setup, find_packages
from jpe_types.setup import run as postSetup
from atexit import register

register(postSetup)

setup(
    name = "jpe_types",
    packages=["jpe_types/paralel",
              "jpe_types",
              "jpe_types/conversions",
              "jpe_types/logging",
              "jpe_types/caching",
              "jpe_types/utils"],
    
    version="0.0.0.4.4",
    license='wtfpl',
    description="slitly improved python types",
    long_description="""s
    # Threading
    -   adds thread inharitance and return posibility
    -   logging filteter for thread inheritance
    -   a thread and a lock in the same class to make it slitly easyer to work with threads

    # conversions
    - adds the abilty to convert ints to any base

    # logging:
    - logging.log() decorator witch logs function calles
    - logging.createDefaultLogger creats a default logger to logg cals curently rather restricted

    # utils:
    log scanning:
    - adds the ability to scan all log files for keywords
    copy:
    - fix fore schalow copy bug in lists see doc
    
    """,
    include_package_data = True,
    author = 'Julian Wandhoven',                   # Type in your name
    author_email = 'julian.wandhoven@fgz.ch',
    install_requires = [
        "cython"
    ],

    url="https://github.com/JulianWww/jpe_types",
    download_url="https://github.com/JulianWww/jpe_types/archive/loggingPathc.tar.gz",
    keywords=["dtypes", "jpe", "utils", "paralel", "logging"],
    classifiers=[
    'Development Status :: 3 - Alpha',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
    'Intended Audience :: Developers',      # Define that your audience are developers
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',   # Again, pick a license
    'Programming Language :: Python :: 3.6' ,
    'Programming Language :: Python :: 3'],#Specify which pyhton versions that you want to support
)



import pickle
import typing
from os.path import join, isdir
from os import mkdir, remove, listdir, rmdir
from shutil import copytree as copyFiles
from shutil import rmtree
from warnings import warn


class NothingAtAll:
    pass

class unpicklableObject(Warning):
    "raised when a value that cant be added to this is added"
    pass

class strCachDir:
    r"""directory with saved items

        a dict with itens and keys saved to file the key must be a string if it isnt keys will be found by the __str__ method
        alsow no \, /, 
        """
    dirFile = None
    lastKey = None
    def __init__(self, inpdict: dict={}):
        """costructor

        create a strCachDir

        @param dir: the inital directory
        @type dir: dict
        """
        self._createDir()
        for key, val in inpdict.items():
            self.__setitem__(key, val)
    
    def _createDir(self):
        "create dir file and save it"
        if not isdir("./cach"):
            mkdir("./cach")
        
        self.dirFile = f"./cach/strDir_{id(self)}"
        if not isdir(self.dirFile):
            mkdir(self.dirFile)
    
    def _isSetup(self):
        "check weather we are setup"
        assert not self.dirFile is None, "setup before this operation"
        assert isdir("./cach"), f"cach may not be deleted"
        assert isdir(self.dirFile), f"nothing in cach may be deleted manualy"

    def _getKeyName(self, key):
        self._isSetup()
        assert hasattr(key, "__str__"), f"key must be str for strCachDir not {type(key)}"
        key = str(key)
        assert not "/" in key, "/ invalid char for key"
        return key

    def __setitem__(self, key, val):
        "add an item"
        key = self._getKeyName(key)

        pickle.dump(val, open(join(self.dirFile, key), "wb"))
        self.lastkey = key

    def __getitem__(self, key):
        key = self._getKeyName(key)

        return pickle.load(open(join(self.dirFile, key), "rb"))
    def get(self, key):
        try:
            return self.__getitem__(key)
        except:
            return
    def pop(self, key):
        val = self.__getitem__(key)
        remove(join(self.dirFile, key))
        return val

    def popitem(self):
        """return the last key"""
        return self.__getitem__(self.lastkey)

    def __str__(self):
        return f"cach dir with keys {self.keys()}"

    def __delitem__(self, key):
        key = self._getKeyName(key)

        remove(join(self.dirFile, key))
    
    def __contains__(self, key):
        key = self._getKeyName(key)
        return key in listdir(self.dirFile)

    def keys(self) -> list:
        return listdir(self.dirFile)
    def __iter__(self):
        return self.keys().__iter__()

    def __copy__(self):
        copied_Dir = strCachDir()
        rmdir(copied_Dir.dirFile)
        copyFiles(self.dirFile, copied_Dir.dirFile)
        return copied_Dir
    
    def __len__(self):
        return len(listdir(self.dirFile))

    def clear(self):
        rmtree(self.dirFile)
        self._createDir()

    def __eq__(self, other):
        if not isinstance(other, (dict, strCachDir)):
            return False
        for myKey, otherKey in zip (self.__iter__(), other):
            if not myKey == otherKey:
                return False
            
            if not self[myKey] == other[otherKey]:
                return False
        return True

    @staticmethod
    def fromkeys(keys):
        return strCachDir(dict.fromkeys(keys))
    
    def items(self)->list:
        out = []
        for key in self:
            out.append((key, self[key]))
        return out

    def setdefault(self, key, val):
        """emm

        The setdefault() method retuns the value of the item with the specified key.
        If the key dose not exist, insert the key, with the specified value, see example below
        
        @param key: the key of the value
        @type key: str
        @param val: the what to set the value to optional
        @type val: any"""
        res = self.get(key)
        if res is None:
            self.__setitem__(key, val)
            res = val
        return res

    def update(self, dic=NothingAtAll, **kwargs) -> None:
        """

        The update() method updates the dictonary with the elements from another dictionary object or from an iterable of key/value paris.

        @param dic: the dict of the value to be returned
        """
        def subscript(**kwargs):
            for key, val in kwargs.items():
                if key == "dic" and val is NothingAtAll:
                    continue
                self[key] = val
            return


        if isinstance(dic, dict):
            return subscript(**dic, **kwargs)
        return subscript(dic=dic, **kwargs) 

    def values(self):
        out = []
        for key in self:
            out.append(self[key])
        return out






"""log decorator

add the log decorator 
"""
from jpe_types.logging import logUtils
from threading import current_thread
import jpe_types.paralel.thread
import os, traceback
from inspect import getfullargspec


main_Logger = None
"std logger"

error_Logger = None
"error used to log errors"

def _setupLoggers():
    """subproces of setup

    dont call directly
    """
    global main_Logger, error_Logger
    main_Logger = logUtils.Create_defaultLogger(logType="mainLogger")
    error_Logger= logUtils.Create_defaultLogger(logType="ErrorLoger", name="ErrorLogger")




def log(log=None, logName=f"{__name__} log", useStdLog=True, loggErrors=True):
    """logging decorator
        
        log function cals

        if the function has a parameter log (must be kwarg) or kwargs the logger used by the decorator to
        log the function call for further logging
            
        @param logger: optional overide logger
        @param logName: the name of the logger when generated
        @param useDtdLog: bool if true will use std logger (main Logger)"""
    if log is None and not useStdLog: 
        log = logUtils.Create_defaultLogger(name=logName)
    def wrap(fun):
        def getMsg(args, kwargs):
            return f"ran function {fun.__name__} with args {args} and kwargs {kwargs}"
        def endExecLog(val):
            return f"function {fun.__name__} execution ended with return val {val}"

        def call(args, kwargs, funLogger):
            data = getfullargspec(fun)
            try:
                if "log" in data[0] or not data[2] is None:
                    val = fun(*args, **kwargs, log=funLogger)
                else:
                    val =fun(*args, **kwargs)
            except Exception as e:
                if loggErrors:
                    for subLoggerStr in logUtils.logging_data["logErors"]:
                        # check weather initiated
                        if eval(subLoggerStr) is None:
                            raise logUtils.LogSetupError()

                        if logUtils.logging_data[logUtils.logging_data["logTranslator"][subLoggerStr]]["trackBack"]:
                            eval(subLoggerStr).error(traceback.format_exc())
                        else:
                            eval(subLoggerStr).error(e)
                        for msg in logUtils.logging_data[logUtils.logging_data["logTranslator"][subLoggerStr]]["errorLogMessage"]:
                            eval(subLoggerStr).error(msg)
                raise e
            funLogger.debug(endExecLog(val))
            return val
            
                    
        def thread_log_wraper(args, kwargs):
            logg = current_thread().threadLogger
            logg.debug(getMsg(args, kwargs))

            return call(args, kwargs, logg)
        
        def log_wraperNoType(args, kwargs):
            if main_Logger is None:
                raise logUtils.LogSetupError
            
            main_Logger.debug(getMsg(args, kwargs))
            return call(args, kwargs, main_Logger)
        

        def log_wraper(args, kwargs):
            log.debug(getMsg(args, kwargs))
            return call(args, kwargs, log)
        
        def wraper(*args, **kwargs):

            thread = current_thread()

            if isinstance(thread, jpe_types.paralel.thread.LogThread):
                return thread_log_wraper(args, kwargs)

            elif useStdLog:
                return log_wraperNoType(args, kwargs)
            
            return log_wraper(args, kwargs)

        return wraper
    return wrap



"utils module"

from jpe_types.utils.copy import *
from jpe_types.utils.logScanner import *

def fallback(last, *fallbacks):
    """decorator for fallback functions

        if the current function retuns an error last will be executed anything after that will be used as a fallback for the one before
        if the last one retuns an error the last error is returned
        @param last: pyobject with __call__ method
        """
    def doWrap(currentFunc):
        def fallback_wraper(*args, **kwargs):
            try:
                return currentFunc(*args, **kwargs)
            except Exception as e:
                if len(fallbacks) == 0:
                    raise e
                return fallback(*fallbacks)(last)(*args, **kwargs)
        return fallback_wraper
    return doWrap
            
        
